@@ -141,10 +141,9 @@ def run_sampling_opencl(G_m_np, thresholds_np, shots, n, is_g_buf_reused):
     dtype = opencl_context.dtype
     wgs = opencl_context.work_group_size
 
-    max_local_size = wgs  # tune
-    max_global_size = ((opencl_context.MAX_GPU_PROC_ELEM + max_local_size - 1) // max_local_size) * max_local_size  # corresponds to MAX_PROC_ELEM macro in OpenCL kernel program
-    global_size = min(((shots + max_local_size - 1) // max_local_size) * max_local_size, max_global_size)
-    local_size = max_local_size
+    local_size = min(wgs, shots)  # tune
+    max_global_size = ((opencl_context.MAX_GPU_PROC_ELEM + local_size - 1) // local_size) * local_size  # corresponds to MAX_PROC_ELEM macro in OpenCL kernel program
+    global_size = min(((shots + local_size - 1) // local_size) * local_size, max_global_size)
     num_groups = global_size // local_size
 
     # Bit-packing params
@@ -276,6 +275,11 @@ def maxcut_tfim(
 
             return "01", weight, ([nodes[0]], [nodes[1]])
 
+    is_segmented = (G_m.nbytes << 1) > opencl_context.max_alloc
+    if is_segmented and is_alt_gpu_sampling:
+        print("[WARN] Using segmented solver, so disabling is_alt_gpu_sampling.")
+        is_alt_gpu_sampling = False
+
     if quality is None:
         quality = 3
 
