@@ -3,6 +3,7 @@ import networkx as nx
 import numpy as np
 import os
 from numba import njit, prange
+from scipy.sparse import lil_matrix
 
 
 @njit
@@ -20,6 +21,37 @@ def get_cut(solution, nodes):
     return bit_string, l, r
 
 
+@njit
+def binary_search(l, t):
+    left = 0
+    right = len(l) - 1
+
+    while left <= right:
+        mid = (left + right) >> 1
+
+        if l[mid] == t:
+            return mid
+
+        if l[mid] < t:
+            left = mid + 1
+        else:
+            right = mid - 1
+
+    return len(l)
+
+
+def to_scipy_sparse_upper_triangular(G, nodes, n_nodes, dtype):
+    lil = lil_matrix((n_nodes, n_nodes), dtype=opencl_context.dtype)
+    for u in range(n_nodes):
+        u_node = nodes[u]
+        for v in range(u + 1, n_nodes):
+            v_node = nodes[v]
+            if G.has_edge(u_node, v_node):
+                lil[u, v] = G[u_node][v_node].get('weight', 1.0)
+
+    return lil.tocsr()
+
+
 @njit(parallel=True)
 def init_theta(h_mult, n_qubits, J_eff, degrees, dtype):
     theta = np.empty(n_qubits, dtype=dtype)
@@ -146,17 +178,20 @@ def probability_by_hamming_weight(J, h, z, theta, t, n_qubits, dtype):
 
 
 class OpenCLContext:
-    def __init__(self, p, a, w, d, c, q, i, m, b, s, k, l):
+    def __init__(self, p, a, w, d, r, c, q, i, m, b, s, x, y, k, l):
         self.MAX_GPU_PROC_ELEM = p
         self.IS_OPENCL_AVAILABLE = a
         self.work_group_size = w
         self.dtype = d
+        self.max_alloc = r
         self.ctx = c
         self.queue = q
         self.init_theta_kernel = i
         self.maxcut_hamming_cdf_kernel = m
         self.bootstrap_kernel = b
         self.bootstrap_sparse_kernel = s
+        self.bootstrap_segmented_kernel = x
+        self.bootstrap_sparse_segmented_kernel = y
         self.sample_for_solution_best_bitset_kernel = k
         self.sample_for_solution_best_bitset_sparse_kernel = l
         self.G_m_buf = None
@@ -170,10 +205,13 @@ queue = None
 compute_units = None
 dtype = np.float32
 work_group_size = 32
+max_alloc = 0xFFFFFFFFFFFFFFFF
 init_theta_kernel = None
 maxcut_hamming_cdf_kernel = None
 bootstrap_kernel = None
 bootstrap_sparse_kernel = None
+bootstrap_segmented_kernel = None
+bootstrap_sparse_segmented_kernel = None
 sample_for_solution_best_bitset_kernel = None
 sample_for_solution_best_bitset_sparse_kernel = None
 
@@ -209,6 +247,9 @@ else:
 
 try:
     import pyopencl as cl
+    import warnings
+
+    warnings.simplefilter("ignore", cl.CompilerWarning)
     
     # Pick a device (GPU if available)
     ctx = cl.create_some_context()
@@ -224,6 +265,8 @@ try:
     maxcut_hamming_cdf_kernel = program.maxcut_hamming_cdf
     bootstrap_kernel = program.bootstrap
     bootstrap_sparse_kernel = program.bootstrap_sparse
+    bootstrap_segmented_kernel = program.bootstrap_segmented
+    bootstrap_sparse_segmented_kernel = program.bootstrap_sparse_segmented
     sample_for_solution_best_bitset_kernel = program.sample_for_solution_best_bitset
     sample_for_solution_best_bitset_sparse_kernel = program.sample_for_solution_best_bitset_sparse
 
@@ -231,8 +274,10 @@ try:
         cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
         ctx.devices[0]
     )
+
+    max_alloc = ctx.devices[0].get_info(cl.device_info.MAX_MEM_ALLOC_SIZE)
 except ImportError:
     IS_OPENCL_AVAILABLE = False
     print("PyOpenCL not installed. (If you have any OpenCL accelerator devices with available ICDs, you might want to optionally install pyopencl.)")
 
-opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, work_group_size, dtype, ctx, queue, init_theta_kernel, maxcut_hamming_cdf_kernel, bootstrap_kernel, bootstrap_sparse_kernel, sample_for_solution_best_bitset_kernel, sample_for_solution_best_bitset_sparse_kernel)
+opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, work_group_size, dtype, max_alloc, ctx, queue, init_theta_kernel, maxcut_hamming_cdf_kernel, bootstrap_kernel, bootstrap_sparse_kernel, bootstrap_segmented_kernel, bootstrap_sparse_segmented_kernel, sample_for_solution_best_bitset_kernel, sample_for_solution_best_bitset_sparse_kernel)
